<?php

    namespace FederationLib;

    use Exception;
    use FederationLib\Classes\Configuration;
    use FederationLib\Enums\Standard\ErrorCodes;
    use FederationLib\Enums\Standard\Methods;
    use FederationLib\Exceptions\DatabaseException;
    use FederationLib\Exceptions\Standard\AccessDeniedException;
    use FederationLib\Exceptions\Standard\ClientNotFoundException;
    use FederationLib\Exceptions\Standard\InternalServerException;
    use FederationLib\Exceptions\Standard\InvalidClientDescriptionException;
    use FederationLib\Exceptions\Standard\InvalidClientNameException;
    use FederationLib\Exceptions\Standard\InvalidFederatedAddressException;
    use FederationLib\Exceptions\Standard\InvalidPeerAssociationTypeException;
    use FederationLib\Exceptions\Standard\InvalidPeerMetadataException;
    use FederationLib\Exceptions\Standard\UnsupportedPeerType;
    use FederationLib\Managers\AssociationManager;
    use FederationLib\Managers\ClientManager;
    use FederationLib\Managers\PeerManager;
    use FederationLib\Objects\ClientRecord;
    use FederationLib\Objects\ResolvedIdentity;
    use FederationLib\Objects\Standard\ClientIdentity;
    use FederationLib\Objects\Standard\PeerUpdate;
    use TamerLib\Enums\TamerMode;
    use TamerLib\tm;
    use Throwable;

    class FederationLib
    {
        /**
         * @var ClientManager
         */
        private ClientManager $client_manager;

        /**
         * @var PeerManager
         */
        private PeerManager $peer_manager;

        /**
         * @var AssociationManager
         */
        private AssociationManager $association_manager;

        /**
         * FederationLib constructor.
         */
        public function __construct()
        {
            $this->client_manager = new ClientManager($this);
            $this->peer_manager = new PeerManager($this);
            $this->association_manager = new AssociationManager($this);
        }

        /**
         * Registers functions to the TamerLib instance, if applicable
         *
         * @return void
         */
        public function registerFunctions(): void
        {
            if(tm::getMode() !== TamerMode::WORKER)
            {
                return;
            }

            $this->client_manager->registerFunctions();
        }

        /**
         * @return string
         */
        public static function getSubprocessPath(): string
        {
            return __DIR__ . DIRECTORY_SEPARATOR . 'subproc';
        }

        /**
         * Resolves the permission role from the given identity and attempts to check if the identity has the
         * required permission to perform the given method
         *
         * @param ClientIdentity|null $identity
         * @throws AccessDeniedException
         * @throws ClientNotFoundException
         * @throws InternalServerException
         * @return ResolvedIdentity
         */
        private function resolveIdentity(?ClientIdentity $identity): ResolvedIdentity
        {
            if($identity === null)
            {
                return new ResolvedIdentity(null, null, true);
            }

            $get_client = tm::do('client_getClient', [$identity->getClientUuid()]);
            $peer = null;

            try
            {
                $client = tm::waitFor($get_client);
                tm::dof('client_updateLastSeen');
            }
            catch(ClientNotFoundException $e)
            {
                tm::clear();
                throw new ClientNotFoundException('Invalid client UUID or client does not exist', $e);
            }
            catch(Exception|Throwable $e)
            {
                tm::clear();
                throw new InternalServerException('There was an error while trying to access the client', $e);
            }

            return new ResolvedIdentity($client, $peer);
        }

        /**
         * Checks if the given identity has the required permission to perform the given method
         *
         * @param string $method
         * @param ResolvedIdentity $resolved_identity
         * @return bool
         */
        private function checkPermission(string $method, ResolvedIdentity $resolved_identity): bool
        {
            return $resolved_identity->getPermissionRole() <= Configuration::getMethodPermission($method);
        }

        /**
         * Pings the client
         *
         * @param ClientIdentity|null $identity
         * @return bool
         * @throws AccessDeniedException
         * @throws ClientNotFoundException
         * @throws InternalServerException
         */
        public function ping(?ClientIdentity $identity): bool
        {
            if(!$this->checkPermission(Methods::PING, $this->resolveIdentity($identity)))
            {
                throw new Exceptions\Standard\AccessDeniedException('You do not have permission to perform this action');
            }

            return true;
        }

        /**
         * @param ClientIdentity|null $identity
         * @return string
         * @throws AccessDeniedException
         * @throws ClientNotFoundException
         * @throws InternalServerException
         */
        public function whoami(?ClientIdentity $identity): string
        {
            $resolved_identity = $this->resolveIdentity($identity);

            if(!$this->checkPermission(Methods::WHOAMI, $resolved_identity))
            {
                throw new Exceptions\Standard\AccessDeniedException('You do not have permission to perform this action');
            }

            if($resolved_identity->getPeer() !== null)
            {
                return $resolved_identity->getPeer()->getAddress();
            }

            if($resolved_identity->getClient() !== null)
            {
                return $resolved_identity->getClient()->getUuid();
            }

            return 'root';
        }

        /**
         * Registers a new client into the database
         *
         * @param ClientIdentity|null $identity
         * @param string|null $name
         * @param string|null $description
         * @return string
         * @throws AccessDeniedException
         * @throws ClientNotFoundException
         * @throws DatabaseException
         * @throws InternalServerException
         * @throws InvalidClientDescriptionException
         * @throws InvalidClientNameException
         */
        public function createClient(?ClientIdentity $identity, ?string $name=null, ?string $description=null): string
        {
            if(!$this->checkPermission(Methods::CREATE_CLIENT, $this->resolveIdentity($identity)))
            {
                throw new Exceptions\Standard\AccessDeniedException('You do not have sufficient permission to create a client');
            }

            try
            {
                return $this->client_manager->registerClient($name, $description);
            }
            catch(Exception $e)
            {
                if(in_array($e->getCode(), ErrorCodes::ALL, true))
                {
                    throw $e;
                }

                throw new Exceptions\Standard\InternalServerException('There was an error while creating the client', $e);
            }
        }

        /**
         * Returns an existing client from the database
         *
         * @param ClientIdentity|null $identity
         * @param string|ClientRecord $client_uuid
         * @return Objects\Standard\Client
         * @throws ClientNotFoundException
         * @throws DatabaseException
         * @throws InternalServerException
         * @throws AccessDeniedException
         */
        public function getClient(?ClientIdentity $identity, string|ClientRecord $client_uuid): Objects\Standard\Client
        {
            if(!$this->checkPermission(Methods::GET_CLIENT, $this->resolveIdentity($identity)))
            {
                throw new Exceptions\Standard\AccessDeniedException('You do not have sufficient permission to fetch a client from the database');
            }

            try
            {
                // Return the standard client object
                return new Objects\Standard\Client($this->client_manager->getClient($client_uuid));
            }
            catch(Exception $e)
            {
                if(in_array($e->getCode(), ErrorCodes::ALL, true))
                {
                    throw $e;
                }

                throw new Exceptions\Standard\InternalServerException('There was an error while getting the client', $e);
            }
        }

        /**
         * Updates the name of an existing client, return True if successful
         *
         * @param ClientIdentity|null $identity
         * @param string $client_uuid
         * @param string $new_name
         * @return bool
         * @throws AccessDeniedException
         * @throws ClientNotFoundException
         * @throws DatabaseException
         * @throws InternalServerException
         * @throws InvalidClientNameException
         */
        public function updateClientName(?ClientIdentity $identity, string $client_uuid, ?string $new_name): bool
        {
            if(!$this->checkPermission(Methods::UPDATE_CLIENT_NAME, $this->resolveIdentity($identity)))
            {
                throw new Exceptions\Standard\AccessDeniedException('You do not have sufficient permission to update the name of a client');
            }

            try
            {
                $this->client_manager->changeClientName($client_uuid, $new_name);
            }
            catch(Exception $e)
            {
                if(in_array($e->getCode(), ErrorCodes::ALL, true))
                {
                    throw $e;
                }

                throw new Exceptions\Standard\InternalServerException('There was an error while updating the client name', $e);
            }

            return true;
        }

        /**
         * Updates the client's description, return True if successful
         *
         * @param ClientIdentity|null $identity
         * @param string $client_uuid
         * @param string|null $new_description
         * @return bool
         * @throws AccessDeniedException
         * @throws ClientNotFoundException
         * @throws DatabaseException // TODO: DatabaseException should not be returning from FederationLib, it's not standard
         * @throws InternalServerException
         * @throws InvalidClientDescriptionException
         */
        public function updateClientDescription(?ClientIdentity $identity, string $client_uuid, ?string $new_description): bool
        {
            if(!$this->checkPermission(Methods::UPDATE_CLIENT_DESCRIPTION, $this->resolveIdentity($identity)))
            {
                throw new Exceptions\Standard\AccessDeniedException('You do not have sufficient permission to update the description of a client');
            }

            try
            {
                $this->client_manager->updateClientDescription($client_uuid, $new_description);
            }
            catch(Exception $e)
            {
                if(in_array($e->getCode(), ErrorCodes::ALL, true))
                {
                    throw $e;
                }

                throw new Exceptions\Standard\InternalServerException('There was an error while updating the client description', $e);
            }

            return true;
        }

        /**
         * Updates the client's permission role, return True if successful
         *
         * @param ClientIdentity|null $identity
         * @param string $client_uuid
         * @param int $new_permission_role
         * @return bool
         * @throws AccessDeniedException
         * @throws ClientNotFoundException
         * @throws DatabaseException
         * @throws Exceptions\Standard\InvalidPermissionRoleException
         * @throws InternalServerException
         */
        public function updateClientPermissionRole(?ClientIdentity $identity, string $client_uuid, int $new_permission_role): bool
        {
            if(!$this->checkPermission(Methods::UPDATE_CLIENT_PERMISSION_ROLE, $this->resolveIdentity($identity)))
            {
                throw new Exceptions\Standard\AccessDeniedException('You do not have sufficient permission to update the permission role of a client');
            }

            try
            {
                $this->client_manager->updateClientPermissionRole($client_uuid, $new_permission_role);
            }
            catch(Exception $e)
            {
                if(in_array($e->getCode(), ErrorCodes::ALL, true))
                {
                    throw $e;
                }

                throw new Exceptions\Standard\InternalServerException('There was an error while updating the client permission role', $e);
            }

            return true;
        }

        /**
         * Syncs the peer to the database, this requires the peer's metadata to be set so that the peer can be
         * registered or updated if it already exists, additionally more information about the peer can be set
         * such as the peer's association.
         *
         * Returns True if successful
         *
         * @param ClientIdentity|null $identity
         * @param PeerUpdate $peer_update
         * @return bool
         * @throws AccessDeniedException
         * @throws ClientNotFoundException
         * @throws DatabaseException
         * @throws InvalidFederatedAddressException
         * @throws InvalidPeerAssociationTypeException
         * @throws InvalidPeerMetadataException
         * @throws UnsupportedPeerType
         * @throws InternalServerException
         */
        public function syncPeer(?ClientIdentity $identity, PeerUpdate $peer_update): bool
        {
            if(!$this->checkPermission(Methods::SYNC_PEER, $this->resolveIdentity($identity)))
            {
                throw new Exceptions\Standard\AccessDeniedException('You do not have sufficient permission to sync a peer');
            }

            if($identity === null)
            {
                throw new Exceptions\Standard\AccessDeniedException('You must be authenticated to sync a peer');
            }

            if($peer_update->getMetadata() === null)
            {
                throw new Exceptions\Standard\InvalidPeerMetadataException('You must provide metadata to sync a peer');
            }

            // TODO: Make this run in parallel

            // Update the metadata if there's any
            if($peer_update->getMetadata() !== null)
            {
                try
                {
                    $this->peer_manager->syncPeer($identity->getClientUuid(), $peer_update->getAddress(), $peer_update->getMetadata());
                }
                catch(Exception $e)
                {
                    if(in_array($e->getCode(), ErrorCodes::ALL, true))
                    {
                        throw $e;
                    }

                    throw new Exceptions\Standard\InternalServerException('There was an error while syncing the peer', $e);
                }
            }

            // Update the association if there's one
            if($peer_update->getAssociation() !== null)
            {
                $association = $peer_update->getAssociation();
                /** @noinspection NullPointerExceptionInspection */
                $this->association_manager->associate($identity->getClientUuid(),
                    $peer_update->getAddress(), $association->getParent(), $association->getType()
                );
            }

            return true;
        }

    }