<?php

    namespace FederationLib\Interfaces;

    use FederationLib\Exceptions\Standard\InvalidPeerMetadataException;
    use FederationLib\Objects\ParsedFederatedAddress;

    interface PeerMetadataInterface extends SerializableObjectInterface
    {
        /**
         * Validates if the metadata is valid or not
         * throws InvalidPeerMetadataException if the metadata is invalid
         *
         * @throws InvalidPeerMetadataException
         */
        public function validate(): void;

        /**
         * Returns True if the metadata indicates that the peer is a bot
         * Returns False if there's no indication that the peer is a bot
         *
         * @return bool
         */
        public function isAutomated(): bool;

        /**
         * Returns the platform-specific ID of the peer (e.g. Telegram Chat ID)
         * If no ID is available, the method will try to find a unique ID in the metadata
         * that can uniquely identify the peer
         *
         * @return string
         */
        public function getPlatformId(): string;

        /**
         * Returns the calculated standard federated address of the peer
         *
         * @return ParsedFederatedAddress
         */
        public function getFederatedAddress(): ParsedFederatedAddress;

        /**
         * Sets the timestamp of when the peer record was last updated (this is usually used internally)
         *
         * @param int|null $timestamp
         * @return void
         */
        public function setLastUpdatedTimestamp(?int $timestamp): void;

        /**
         * Returns the timestamp of when the peer record was last updated
         *
         * @return int|null
         */
        public function getLastUpdatedTimestamp(): ?int;

        /**
         * Sets the Client's UUID that last updated the peer record (this is usually used internally)
         *
         * @param string|null $client_uuid
         * @return void
         */
        public function setUpdatedClient(?string $client_uuid): void;

        /**
         * Returns the Client's UUID that last updated the peer record
         *
         * @return string|null
         */
        public function getUpdatedClient(): ?string;

        /**
         * Returns an array representation of the object, if $raw is true, the array will contain
         * additional information that is not part of the standard metadata such as the client UUID
         * and the last updated timestamp
         *
         * @param bool $raw
         * @return array
         */
        public function toArray(bool $raw=false): array;

        /**
         * Constructs object from an array representation, if $raw is enabled it will accept
         * raw data that is not part of the standard metadata, this should be disabled if
         * the data is coming from an untrusted source such as a user
         *
         * @param array $array
         * @param bool $raw
         * @return PeerMetadataInterface
         */
        public static function fromArray(array $array, bool $raw=false): PeerMetadataInterface;
    }