<?php

    namespace FederationLib\Managers;

    use Doctrine\DBAL\ParameterType;
    use Exception;
    use FederationLib\Classes\Configuration;
    use FederationLib\Classes\Database;
    use FederationLib\Classes\Security;
    use FederationLib\Classes\Utilities;
    use FederationLib\Classes\Validate;
    use FederationLib\Enums\DatabaseTables;
    use FederationLib\Enums\FilterOrder;
    use FederationLib\Enums\Filters\ListClientsFilter;
    use FederationLib\Enums\Misc;
    use FederationLib\Exceptions\DatabaseException;
    use FederationLib\Exceptions\Standard\ClientNotFoundException;
    use FederationLib\Exceptions\Standard\InvalidClientDescriptionException;
    use FederationLib\Exceptions\Standard\InvalidClientNameException;
    use FederationLib\Exceptions\Standard\InvalidPermissionRoleException;
    use FederationLib\FederationLib;
    use FederationLib\Objects\ClientRecord;
    use LogLib\Log;
    use Symfony\Component\Uid\Uuid;
    use TamerLib\Enums\TamerMode;
    use TamerLib\tm;

    class ClientManager
    {
        /**
         * @var FederationLib
         */
        private FederationLib $federationLib;

        /**
         * ClientManager constructor.
         *
         * @param FederationLib $federationLib
         */
        public function __construct(FederationLib $federationLib)
        {
            $this->federationLib = $federationLib;
        }

        /**
         * Registers functions to the TamerLib instance, if applicable
         *
         * @return void
         */
        public function registerFunctions(): void
        {
            if(tm::getMode() !== TamerMode::WORKER)
            {
                return;
            }

            tm::addFunction('client_registerClient', [$this, 'registerClient']);
            tm::addFunction('client_getClient', [$this, 'getClient']);
            tm::addFunction('client_changeClientName', [$this, 'changeClientName']);
            tm::addFunction('client_changeClientDescription', [$this, 'updateClientDescription']);
            tm::addFunction('client_changeClientPermissionRole', [$this, 'updateClientPermissionRole']);
            tm::addFunction('client_updateLastSeen', [$this, 'updateLastSeen']);
            tm::addFunction('client_listClients', [$this, 'listClients']);
            tm::addFunction('client_getTotalClients', [$this, 'getTotalClients']);
            tm::addFunction('client_getTotalPages', [$this, 'getTotalPages']);
            tm::addFunction('client_deleteClient', [$this, 'deleteClient']);
        }

        /**
         * Registers a client into the database, returns the UUID that was generated for the client.
         *
         * @param string|null $name
         * @param string|null $description
         * @return string
         * @throws DatabaseException
         * @throws InvalidClientDescriptionException
         * @throws InvalidClientNameException
         */
        public function registerClient(?string $name=null, ?string $description=null): string
        {
            $qb = Database::getConnection()->createQueryBuilder();
            $qb->insert(DatabaseTables::CLIENTS);
            $uuid = Uuid::v4()->toRfc4122();

            if($name === null)
            {
                $name = Utilities::generateName(4);
            }

            if(!Validate::clientName($name))
            {
                throw new InvalidClientNameException(sprintf('Invalid client name: %s', $name));
            }

            if($description !== null && strlen($description) > 128)
            {
                throw new InvalidClientDescriptionException(sprintf('Invalid client description: %s', $description));
            }

            $qb->setValue('uuid', $qb->createNamedParameter($uuid));
            $qb->setValue('name', $qb->createNamedParameter($name));
            $qb->setValue('description', $qb->createNamedParameter($description, (is_null($description) ? ParameterType::NULL : ParameterType::STRING)));
            $qb->setValue('secret_totp', $qb->createNamedParameter(Security::generateSecret()));
            $qb->setValue('flags', $qb->createNamedParameter(null, ParameterType::NULL));

            try
            {
                $qb->executeStatement();
            }
            catch(Exception $e)
            {
                throw new DatabaseException('Failed to register client: ' . $e->getMessage(), $e);
            }

            Log::info(Misc::FEDERATIONLIB, sprintf('Registered client with UUID %s', $uuid));
            return $uuid;
        }

        /**
         * Returns an existing client from the database.
         *
         * @param string|ClientRecord $client_uuid
         * @return ClientRecord
         * @throws ClientNotFoundException
         * @throws DatabaseException
         */
        public function getClient(string|ClientRecord $client_uuid): ClientRecord
        {
            if($client_uuid instanceof ClientRecord)
            {
                $client_uuid = $client_uuid->getUuid();
            }

            // Use the cache first if it's enabled
            if(Configuration::isCacheSystemEnabled() && Configuration::getObjectCacheEnabled('client_objects'))
            {
                try
                {
                    $redis = RedisConnectionManager::getConnectionFromConfig('client_objects');

                    if($redis->exists($client_uuid))
                    {
                        $client = ClientRecord::fromArray($redis->hGetAll($client_uuid));
                        Log::debug(Misc::FEDERATIONLIB, sprintf('Loaded client object %s from cache', $client_uuid));
                        return $client;
                    }
                }
                catch(Exception $e)
                {
                    Log::warning(Misc::FEDERATIONLIB, sprintf('Failed to load client object %s from cache: %s', $client_uuid, $e->getMessage()));
                }
            }

            $qb = Database::getConnection()->createQueryBuilder();
            $qb->select('*');
            $qb->from(DatabaseTables::CLIENTS);
            $qb->where('uuid = :uuid');
            $qb->setParameter('uuid', $client_uuid);
            $qb->setMaxResults(1);

            try
            {
                $result = $qb->executeQuery();

                if($result->rowCount() === 0)
                {
                    throw new ClientNotFoundException($client_uuid);
                }

                $client = ClientRecord::fromArray($result->fetchAssociative());
            }
            catch(ClientNotFoundException $e)
            {
                throw $e;
            }
            catch(Exception $e)
            {
                throw new DatabaseException('Failed to get Client: ' . $e->getMessage(), $e);
            }

            // Store the record in the cache if caching is enabled.
            if(Configuration::isCacheSystemEnabled() && Configuration::getObjectCacheEnabled('client_objects'))
            {
                try
                {
                    if(!isset($redis))
                    {
                        $redis = RedisConnectionManager::getConnectionFromConfig('client_objects');
                    }

                    $redis->hMSet($client->getUuid(), $client->toArray());
                    if(Configuration::getObjectCacheTTL('client_objects') > 0)
                    {
                        $redis->expire($client->getUuid(), Configuration::getObjectCacheTTL('client_objects'));
                    }

                    Log::debug(Misc::FEDERATIONLIB, sprintf('Cached client object %s', $client->getUuid()));
                }
                catch(Exception $e)
                {
                    Log::warning(Misc::FEDERATIONLIB, sprintf('Failed to cache client object %s: %s', $client->getUuid(), $e->getMessage()), $e);
                }
            }

            return $client;
        }

        /**
         * Changes the name of a client.
         *
         * @param string|ClientRecord $client_uuid
         * @param string|null $name
         * @return void
         * @throws ClientNotFoundException
         * @throws DatabaseException
         * @throws InvalidClientNameException
         */
        public function changeClientName(string|ClientRecord $client_uuid, ?string $name=null): void
        {
            if($client_uuid instanceof ClientRecord)
            {
                $client_uuid = $client_uuid->getUuid();
            }

            if($name === null)
            {
                $name = Utilities::generateName(4);
            }

            if(!Validate::clientName($name))
            {
                throw new InvalidClientNameException(sprintf('Invalid client name: %s', $name));
            }

            $update_timestamp = time();
            $qb = Database::getConnection()->createQueryBuilder();
            $qb->update(DatabaseTables::CLIENTS);
            $qb->set('name', ':name');
            $qb->setParameter('name', $name);
            $qb->set('updated_timestamp', ':updated_timestamp');
            $qb->setParameter('updated_timestamp', $update_timestamp, ParameterType::INTEGER);
            $qb->where('uuid = :uuid');
            $qb->setParameter('uuid', $client_uuid);
            $qb->setMaxResults(1);

            try
            {
                $affected_rows = $qb->executeStatement();
            }
            catch(Exception $e)
            {
                throw new DatabaseException('Failed to change client name: ' . $e->getMessage(), $e);
            }

            if($affected_rows === 0)
            {
                throw new ClientNotFoundException($client_uuid);
            }

            Log::verbose(Misc::FEDERATIONLIB, sprintf('Changed client name for client %s to %s', $client_uuid, $name));

            // Update the record in redis if caching is enabled
            if(Configuration::isCacheSystemEnabled() && Configuration::getObjectCacheEnabled('client_objects'))
            {
                try
                {
                    $redis = RedisConnectionManager::getConnectionFromConfig('client_objects');

                    if(!$redis->exists($client_uuid))
                    {
                       return;
                    }

                    $redis->hSet($client_uuid, 'name', $name);
                    $redis->hSet($client_uuid, 'updated_timestamp', $update_timestamp);
                    if(Configuration::getObjectCacheTTL('client_objects') > 0)
                    {
                        $redis->expire($client_uuid, Configuration::getObjectCacheTTL('client_objects'));
                    }

                    Log::debug(Misc::FEDERATIONLIB, sprintf('Updated client object %s <%s> in cache', $client_uuid, 'name'));

                }
                catch(Exception $e)
                {
                    Log::warning(Misc::FEDERATIONLIB, sprintf('Failed to update client object %s in cache: %s', $client_uuid, $e->getMessage()), $e);
                }
            }
        }

        /**
         * Changes the description of a client
         *
         * @param string|ClientRecord $client_uuid
         * @param string|null $description
         * @return void
         * @throws ClientNotFoundException
         * @throws DatabaseException
         * @throws InvalidClientDescriptionException
         */
        public function updateClientDescription(string|ClientRecord $client_uuid, ?string $description=null): void
        {
            if($client_uuid instanceof ClientRecord)
            {
                $client_uuid = $client_uuid->getUuid();
            }

            if($description !== null && strlen($description) > 128)
            {
                throw new InvalidClientDescriptionException(sprintf('Invalid client description: %s', $description));
            }

            $updated_timestamp = time();
            $qb = Database::getConnection()->createQueryBuilder();
            $qb->update(DatabaseTables::CLIENTS);
            $qb->set('description', ':description');
            $qb->setParameter('description', $description, (is_null($description) ? ParameterType::NULL : ParameterType::STRING));
            $qb->set('updated_timestamp', ':updated_timestamp');
            $qb->setParameter('updated_timestamp', $updated_timestamp, ParameterType::INTEGER);
            $qb->where('uuid = :uuid');
            $qb->setParameter('uuid', $client_uuid);
            $qb->setMaxResults(1);

            try
            {
                $affected_rows = $qb->executeStatement();
            }
            catch(Exception $e)
            {
                throw new DatabaseException('Failed to change client description: ' . $e->getMessage(), $e);
            }

            if($affected_rows === 0)
            {
                throw new ClientNotFoundException($client_uuid);
            }

            Log::verbose(Misc::FEDERATIONLIB, sprintf('Changed client description for client %s to %s', $client_uuid, $description));

            if(Configuration::isCacheSystemEnabled() && Configuration::getObjectCacheEnabled('client_objects'))
            {
                try
                {
                    $redis = RedisConnectionManager::getConnectionFromConfig('client_objects');

                    if(!$redis->exists($client_uuid))
                    {
                        return;
                    }

                    $redis->hSet($client_uuid, 'description', $description);
                    $redis->hSet($client_uuid, 'updated_timestamp', $updated_timestamp);

                    if(Configuration::getObjectCacheTTL('client_objects') > 0)
                    {
                        $redis->expire($client_uuid, Configuration::getObjectCacheTTL('client_objects'));
                    }

                    Log::debug(Misc::FEDERATIONLIB, sprintf('Updated client object %s <%s> in cache', $client_uuid, 'description'));
                }
                catch(Exception $e)
                {
                    Log::warning(Misc::FEDERATIONLIB, sprintf('Failed to update client object %s in cache: %s', $client_uuid, $e->getMessage()), $e);
                }
            }
        }

        /**
         * Updates the permission role of a client.
         *
         * @param string|ClientRecord $client_uuid
         * @param int $permission_role
         * @return void
         * @throws ClientNotFoundException
         * @throws DatabaseException
         * @throws InvalidPermissionRoleException
         */
        public function updateClientPermissionRole(string|ClientRecord $client_uuid, int $permission_role): void
        {
            if($client_uuid instanceof  ClientRecord)
            {
                $client_uuid = $client_uuid->getUuid();
            }

            if(!Validate::permissionRole($permission_role))
            {
                throw new InvalidPermissionRoleException(sprintf('Invalid permission role: %s', $permission_role));
            }

            $updated_timestamp = time();
            $qb = Database::getConnection()->createQueryBuilder();
            $qb->update(DatabaseTables::CLIENTS);
            $qb->set('permission_role', ':permission_role');
            $qb->setParameter('permission_role', $permission_role, ParameterType::INTEGER);
            $qb->set('updated_timestamp', ':updated_timestamp');
            $qb->setParameter('updated_timestamp', $updated_timestamp, ParameterType::INTEGER);
            $qb->where('uuid = :uuid');
            $qb->setParameter('uuid', $client_uuid);
            $qb->setMaxResults(1);

            try
            {
                $affected_rows = $qb->executeStatement();
            }
            catch(Exception $e)
            {
                throw new DatabaseException('Failed to change client permission role: ' . $e->getMessage(), $e);
            }

            if($affected_rows === 0)
            {
                throw new ClientNotFoundException($client_uuid);
            }

            Log::verbose(Misc::FEDERATIONLIB, sprintf('Changed client permission role for client %s to %s', $client_uuid, $permission_role));

            if(Configuration::isCacheSystemEnabled() && Configuration::getObjectCacheEnabled('client_objects'))
            {
                try
                {
                    $redis = RedisConnectionManager::getConnectionFromConfig('client_objects');

                    if(!$redis->exists($client_uuid))
                    {
                        return;
                    }

                    $redis->hSet($client_uuid, 'permission_role', $permission_role);
                    $redis->hSet($client_uuid, 'updated_timestamp', $updated_timestamp);

                    if(Configuration::getObjectCacheTTL('client_objects') > 0)
                    {
                        $redis->expire($client_uuid, Configuration::getObjectCacheTTL('client_objects'));
                    }

                    Log::debug(Misc::FEDERATIONLIB, sprintf('Updated client object %s <%s> in cache', $client_uuid, 'permission_role'));
                }
                catch(Exception $e)
                {
                    Log::warning(Misc::FEDERATIONLIB, sprintf('Failed to update client object %s in cache: %s', $client_uuid, $e->getMessage()), $e);
                }
            }
        }

        /**
         * Updates a client's last seen timestamp.
         *
         * @param string|ClientRecord $uuid
         * @return void
         * @throws DatabaseException
         * @noinspection PhpUnused
         */
        public function updateLastSeen(string|ClientRecord $uuid): void
        {
            if($uuid instanceof ClientRecord)
            {
                $uuid = $uuid->getUuid();
            }

            $timestamp = time();

            $qb = Database::getConnection()->createQueryBuilder();
            $qb->update(DatabaseTables::CLIENTS);
            $qb->set('seen_timestamp', ':seen_timestamp');
            $qb->setParameter('seen_timestamp', $timestamp);
            $qb->where('uuid = :uuid');
            $qb->setParameter('uuid', $uuid);
            $qb->setMaxResults(1);

            try
            {
                $qb->executeStatement();
            }
            catch(Exception $e)
            {
                throw new DatabaseException('Failed to update last seen timestamp: ' . $e->getMessage(), $e);
            }

            Log::verbose(Misc::FEDERATIONLIB, sprintf('Updated last seen timestamp for client %s to %s', $uuid, $timestamp));

            if(Configuration::isCacheSystemEnabled() && Configuration::getObjectCacheEnabled('client_objects'))
            {
                try
                {
                    $redis = RedisConnectionManager::getConnectionFromConfig('client_objects');

                    if(!$redis->exists($uuid))
                    {
                        return;
                    }

                    $redis->hSet($uuid, 'seen_timestamp', $timestamp);

                    if(Configuration::getObjectCacheTTL('client_objects') > 0)
                    {
                        $redis->expire($uuid, Configuration::getObjectCacheTTL('client_objects'));
                    }

                    Log::debug(Misc::FEDERATIONLIB, sprintf('Updated client object %s <%s> in cache', $uuid, 'seen_timestamp'));
                }
                catch(Exception $e)
                {
                    Log::warning(Misc::FEDERATIONLIB, sprintf('Failed to update client object %s in cache: %s', $uuid, $e->getMessage()), $e);
                }
            }
        }

        /**
         * Returns an array of clients based on the filter, order, and page.
         *
         * @param int $page
         * @param string $filter
         * @param string $order
         * @param int $max_items
         * @return array
         * @throws DatabaseException
         */
        public function listClients(int $page, string $filter=ListClientsFilter::CREATED_TIMESTAMP, string $order=FilterOrder::ASCENDING, int $max_items=100): array
        {
            $qb = Database::getConnection()->createQueryBuilder();
            $qb->select(
                'uuid', 'enabled', 'name', 'description', 'secret_totp', 'permission_role',
                'flags', 'created_timestamp', 'updated_timestamp', 'seen_timestamp'
            );

            $qb->from(DatabaseTables::CLIENTS);
            $qb->setFirstResult(($page - 1) * $max_items);
            $qb->setMaxResults($max_items);

            if($order !== FilterOrder::ASCENDING && $order !== FilterOrder::DESCENDING)
            {
                throw new DatabaseException('Invalid order: ' . $order);
            }

            $redis_client = null;

            if(Configuration::isCacheSystemEnabled() && Configuration::getObjectCacheEnabled('client_objects'))
            {
                try
                {
                    $redis_client = RedisConnectionManager::getConnectionFromConfig('client_objects');
                }
                catch(Exception $e)
                {
                    Log::warning(Misc::FEDERATIONLIB, sprintf('Failed to connect to Redis server: %s', $e->getMessage()), $e);
                }
            }

            switch($filter)
            {
                case ListClientsFilter::CREATED_TIMESTAMP:
                    $qb->orderBy('created_timestamp', strtoupper($order));
                    break;

                case ListClientsFilter::UPDATED_TIMESTAMP:
                    $qb->orderBy('updated_timestamp', strtoupper($order));
                    break;

                case ListClientsFilter::SEEN_TIMESTAMP:
                    $qb->orderBy('seen_timestamp', strtoupper($order));
                    break;

                case ListClientsFilter::ENABLED:
                    $qb->orderBy('enabled', strtoupper($order));
                    break;

                default:
                    throw new DatabaseException('Invalid filter: ' . $filter);
            }

            try
            {
                $result = $qb->executeQuery();
                $clients = [];

                while($row = $result->fetchAssociative())
                {
                    $client_object = ClientRecord::fromArray($row);

                    if($redis_client !== null)
                    {
                        $redis_client->hMSet($client_object->getUuid(), $client_object->toArray());

                        if(Configuration::getObjectCacheTTL('client_objects') > 0)
                        {
                            $redis_client->expire($row['uuid'], Configuration::getObjectCacheTTL('client_objects'));
                        }
                    }

                    $clients[] = $client_object;
                }
            }
            catch(Exception $e)
            {
                throw new DatabaseException('Failed to list clients: ' . $e->getMessage(), $e);
            }

            unset($client);
            return $clients;
        }

        /**
         * Returns the total number of clients.
         *
         * @return int
         * @throws DatabaseException
         */
        public function getTotalClients(): int
        {
            $qb = Database::getConnection()->createQueryBuilder();
            $qb->select('COUNT(uuid)');
            $qb->from(DatabaseTables::CLIENTS);

            try
            {
                $result = $qb->executeQuery();
                $row = $result->fetchAssociative();
            }
            catch(Exception $e)
            {
                throw new DatabaseException('Failed to get total clients: ' . $e->getMessage(), $e);
            }

            return (int)$row['COUNT(uuid)'];
        }

        /**
         * Returns teh total number of pages.
         *
         * @param int $max_items
         * @return int
         * @throws DatabaseException
         */
        public function getTotalPages(int $max_items=100): int
        {
            return (int)ceil($this->getTotalClients() / $max_items);
        }

        /**
         * Deletes an existing client from the database.
         *
         * @param string|ClientRecord $uuid
         * @return void
         * @throws DatabaseException
         */
        public function deleteClient(string|ClientRecord $uuid): void
        {
            if($uuid instanceof ClientRecord)
            {
                $uuid = $uuid->getUuid();
            }

            $qb = Database::getConnection()->createQueryBuilder();
            $qb->delete(DatabaseTables::CLIENTS);
            $qb->where('uuid = :uuid');
            $qb->setParameter('uuid', $uuid);
            $qb->setMaxResults(1);

            try
            {
                $qb->executeStatement();
            }
            catch(Exception $e)
            {
                throw new DatabaseException('Failed to delete client: ' . $e->getMessage(), $e);
            }

            if(Configuration::isCacheSystemEnabled() && Configuration::getObjectCacheEnabled('client_objects'))
            {
                try
                {
                    $redis = RedisConnectionManager::getConnectionFromConfig('client_objects');

                    if($redis->exists($uuid))
                    {
                        $redis->del($uuid);
                    }
                }
                catch(Exception $e)
                {
                    Log::warning(Misc::FEDERATIONLIB, sprintf('Failed to delete client object %s from cache: %s', $uuid, $e->getMessage()), $e);
                }
            }
        }
    }