<?php

    namespace FederationLib\Managers;

    use Exception;
    use FederationLib\Classes\Configuration;
    use FederationLib\Classes\Redis;
    use FederationLib\Classes\Utilities;
    use FederationLib\Exceptions\CacheConnectionException;
    use FederationLib\Exceptions\CacheDriverException;
    use InvalidArgumentException;

    class RedisConnectionManager
    {
        /**
         * @var Redis[]|null
         */
        private static $connections;

        /**
         * @param string $name
         * @return \Redis
         * @throws CacheConnectionException
         */
        public static function getConnectionFromConfig(string $name): \Redis
        {
            return self::getConnection(
                Configuration::getObjectCacheServerPreference($name),
                Configuration::getObjectCacheServerFallback($name)
            );
        }

        /**
         * Returns the requested redis connection to use
         *
         * WARNING: Very advanced, terrifying code ahead - proceed with caution
         * This code is responsible for determining which redis connection to use
         * based on the configuration and the state of the connections
         * It's a bit of a mess, but it works
         *
         * @param string|null $name
         * @param string|null $fallback
         * @return \Redis
         * @throws CacheConnectionException
         */
        public static function getConnection(?string $name=null, ?string $fallback=null): \Redis
        {
            if(self::$connections === null)
            {
                self::$connections = [];
                foreach(Configuration::getCacheServers() as $configuration)
                {
                    self::$connections[$configuration->getName()] = new Redis($configuration);
                }
            }

            // If the name isn't null or "any", return the connection with that name
            if($name !== null || strtolower($name) !== 'any')
            {
                if(!isset(self::$connections[$name]))
                {
                    if($fallback !== null)
                    {
                        return self::getConnection($fallback);
                    }

                    throw new InvalidArgumentException("Redis connection with name '$name' not found");
                }

                if(!self::$connections[$name]->isAvailable())
                {
                    if($fallback !== null)
                    {
                        return self::getConnection($fallback);
                    }

                    throw new CacheConnectionException("Redis connection with name '$name' is not available");
                }

                try
                {
                    return self::$connections[$name]->getConnection();
                }
                catch(Exception $e)
                {
                    if($fallback !== null)
                    {
                        return self::getConnection($fallback);
                    }

                    throw new CacheConnectionException(sprintf("Failed to retrieve the connection for \'%s\'", $name), 0, $e);
                }
            }

            // Assuming we're here, we're looking for any connection

            // Build the weights array
            $weights = [];
            /** @var Redis $connection */
            foreach(self::$connections as $connection)
            {
                if($connection->isAvailable())
                {
                    $priority = $connection->getConfiguration()->getPriority();

                    // Calculate the priority based on the connection state and the configuration
                    if(Configuration::getCacheErrorConnectionPriority() !== 0 && $connection->isConnectionError())
                    {
                        $priority = ((Configuration::getCacheErrorConnectionPriority()) + ($priority));
                    }
                    elseif(Configuration::getCacheOpenedConnectionPriority() !== 0 && $connection->isConnected())
                    {
                        $priority = ((Configuration::getCacheOpenedConnectionPriority()) + ($priority));
                    }

                    if((int)$priority > 100)
                    {
                        $priority = 100;
                    }
                    elseif((int)$priority < 0)
                    {
                        $priority = 0;
                    }

                    $weights[$connection->getConfiguration()->getName()] = (int)$priority;
                }
            }

            if(count($weights) === 0)
            {
                // If there are no available connections, this may be based off of the configuration
                // In this case, resort to the default.
                /** @var Redis $connection */
                foreach(self::$connections as $connection)
                {
                    if($connection->isAvailable())
                    {
                        $weights[$connection->getConfiguration()->getName()] = $connection->getConfiguration()->getPriority();
                    }
                }

                // If there are still no available connections, throw an exception
                // It's clearly the user's fault at this point lol
                if (count($weights) === 0)
                {
                    if($fallback !== null)
                    {
                        return self::getConnection($fallback);
                    }

                    throw new CacheConnectionException("No available Redis connections");
                }

                if(count($weights) === 1)
                {
                    // If there's only one available connection, just use that lmao ez
                    return self::$connections[array_key_first($weights)]->getConnection();
                }
            }
            elseif(count($weights) === 1)
            {
                // Same as above
                return self::$connections[array_key_first($weights)]->getConnection();
            }

            $selected_connection = Utilities::weightedRandomPick($weights);

            try
            {
                return self::$connections[$selected_connection]->getConnection();
            }
            catch(Exception $e)
            {
                if($fallback !== null)
                {
                    // After all that, at least we have a fallback
                    return self::getConnection($fallback);
                }

            }
            finally
            {
                // Or not :(
                throw new CacheConnectionException(sprintf("Failed to retrieve the connection for \'%s\'", $selected_connection), 0, $e);
            }

            // Voila! je suis fier de ça
        }
    }