<?php

    namespace FederationLib\Objects\Standard;

    use FederationLib\Classes\Security;
    use FederationLib\Classes\Validate;
    use FederationLib\Exceptions\Standard\InvalidDataException;
    use FederationLib\Interfaces\SerializableObjectInterface;
    use FederationLib\Interfaces\ValidateInterface;
    use FederationLib\Objects\ParsedFederatedAddress;
    use FederationLib\Objects\Standard\PeerUpdate\Association;
    use InvalidArgumentException;

    class PeerUpdate implements SerializableObjectInterface, ValidateInterface
    {
        /**
         * @var string
         */
        private $address;

        /**
         * @var array|null
         */
        private $metadata;

        /**
         * @var Association|null
         */
        private $association;

        /**
         * Validates the given object data, throws an exception if invalid.
         *
         * @return void
         * @throws InvalidDataException
         */
        public function validate(): void
        {
            if(!is_string($this->address))
            {
                throw new InvalidDataException(sprintf('Peer\'s address must be a string, %s given.', Security::gettype($this->address)));
            }

            try
            {
                $parsed_address = new ParsedFederatedAddress($this->address);
            }
            catch(InvalidArgumentException $e)
            {
                throw new InvalidDataException(sprintf('Peer\'s address must be a valid federated address, %s given.', $this->address), $e);
            }

            if(!Validate::validateEntityType($parsed_address->getPeerType()))
            {
                throw new InvalidDataException(sprintf('Peer\'s address must be a valid peer type, %s given.', $parsed_address->getPeerType()));
            }

            if(!is_null($this->metadata) && !is_array($this->metadata))
            {
                throw new InvalidDataException(sprintf('Peer\'s metadata must be an array or null, %s given.', Security::gettype($this->metadata)));
            }

            if(!is_null($this->association))
            {
                $this->association->validate();
            }
        }

        /**
         * Returns the federated address of the peer.
         *
         * @return string
         */
        public function getAddress(): string
        {
            return $this->address;
        }

        /**
         * Optional. Returns the metadata of the peer, or null if none.
         *
         * @return array|null
         */
        public function getMetadata(): ?array
        {
            return $this->metadata;
        }

        /**
         * Optional. Returns the association of the peer, or null if none.
         *
         * @return Association|null
         */
        public function getAssociation(): ?Association
        {
            return $this->association;
        }

        /**
         * Returns an array representation of the string
         *
         * @return array
         */
        public function toArray(): array
        {
            $association = null;

            if($association instanceof Association)
            {
                $association = $this->association->toArray();
            }

            return [
                'address' => $this->address,
                'metadata' => $this->metadata,
                'association' => $association
            ];
        }

        /**
         * Constructs the object from an array representation
         *
         * @param array $array
         * @return PeerUpdate
         */
        public static function fromArray(array $array): PeerUpdate
        {
            $object = new self();

            $object->address = $array['address'];
            $object->metadata = $array['metadata'];

            if(isset($array['association']))
            {
                $object->association = Association::fromArray($array['association']);
            }

            return $object;
        }
    }