<?php

    /** @noinspection PhpMissingFieldTypeInspection */

    namespace FederationLib\Objects\Standard\PeerUpdate;

    use FederationLib\Classes\Security;
    use FederationLib\Classes\Validate;
    use FederationLib\Exceptions\Standard\InvalidDataException;
    use FederationLib\Interfaces\SerializableObjectInterface;
    use FederationLib\Interfaces\ValidateInterface;
    use FederationLib\Objects\ParsedFederatedAddress;
    use InvalidArgumentException;

    class Association implements SerializableObjectInterface, ValidateInterface
    {
        /**
         * @var string
         */
        private $parent;

        /**
         * @var string
         */
        private $type;

        /**
         * Validates the given object data, throws an exception if invalid.
         *
         * @return void
         * @throws InvalidDataException
         */
        public function validate(): void
        {
            if(!is_string($this->parent))
            {
                throw new InvalidDataException(sprintf('Peer\'s assocation parent must be a string, %s given.', Security::gettype($this->parent)));
            }

            if(!is_string($this->type))
            {
                throw new InvalidDataException(sprintf('Peer\'s assocation type must be a string, %s given.', Security::gettype($this->type)));
            }

            try
            {
                $parsed_address = new ParsedFederatedAddress($this->parent);
            }
            catch(InvalidArgumentException $e)
            {
                throw new InvalidDataException(sprintf('Peer\'s association parent must be a valid federated address, %s given.', $this->parent), $e);
            }

            if(!Validate::validateEntityType($parsed_address->getPeerType()))
            {
                throw new InvalidDataException(sprintf('Peer\'s association parent must be a valid peer type, %s given.', $parsed_address->getPeerType()));
            }
        }

        /**
         * Returns the parent of the association.
         *
         * @return string
         */
        public function getParent(): string
        {
            return $this->parent;
        }

        /**
         * Returns the type of association.
         *
         * @return string
         */
        public function getType(): string
        {
            return $this->type;
        }

        /**
         * Returns an array representation of the object.
         *
         * @return array
         */
        public function toArray(): array
        {
            return [
                'parent' => $this->parent,
                'type' => $this->type,
            ];
        }

        /**
         * Constructs the object from an array representation.
         *
         * @param array $array
         * @return Association
         */
        public static function fromArray(array $array): Association
        {
            $object = new self();

            $object->parent = $array['parent'];
            $object->type = $array['type'];

            return $object;
        }
    }